/*
   LibIntelCounters -- Support for reading Intel processor Counter
   Copyright (C) 2010 Leonardo de Paula Rosa Piga

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General
   Public License along with this library; if not, write to the
   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301 USA
*/

#include <unistd.h>

// Kernel syscall numbers
/** Kernel syscall number for rdmsr instruction execution */
#define __NR_rdmsr 340

/** Kernel syscall number for wrmsr instruction execution */
#define __NR_wrmsr 339

// Fixed counters
/* Instr_Retired.Any: */
#if defined(CORE2)
#define FIXED_COUNTER_CONTROL 0x38D

#define PERF_COUNTER_GLOBAL_STATUS 0x38E
#define PERF_COUNTER_CONTROL 0x38F
#define PERF_COUNTER_OVF 0x390

#define INSTR_RETIRED_ANY_COUNTER 0x309
#define CPU_CLK_UNHALTED_CORE     0x30A

#define ENABLE_INSTR_RETIRED_ANY_COUNTER_USR 0x1
#define ENABLE_INSTR_RETIRED_ANY_COUNTER_OS 0x2

#endif

#ifdef CORE2
// Intel Counter Masks
#define EVENT_MASK     0x000000FF // Event maks
#define UMASK_MASK     0x0000FF00 // Umask mask
#define USERC_MASK     0x00010000 // User count flag
#define OSC_MASK       0x00020000 // OS count flag  
#define EDGE_MASK      0x00040000 // Edge enable flag
#define PC_MASK        0x00080000 // Pin Count flag  
#define INT_MASK       0x00100000 // Interruption Flag
#define ANYT_MASK      0x00200000 // Any thread flag  
#define ENB_MASK       0x00400000 // Enable flag      
#define INVERT_MASK    0x00400000 // Invert the CMASKw
#define CMASK_MASK     0xFF000000 // CMASK MASK field 

#define INSTR_COUNTER  (0x000000C0 | EDGE_MASK | PC_MASK | INT_MASK | ANYT_MASK)

#define CYCLES_CONTROL_MASK 0x30
#define CYCLES_CONTROL_USER_COUNT 0x20
#define CYCLES_CONTROL_SYS_COUNT 0x10

// Event selector registers
#define PERF_EVT_SEL0  0x186
#define PERF_EVT_SEL1  0x187

#define GENERAL_PERF_COUNTER0 0xC1
#define GENERAL_PERF_COUNTER1 0xC2

#define ENABLE_PMC0 0x1
#define ENABLE_PMC1 0x2
#define ENABLE_CPU_CLK_UNHALTED_CORE 0x2

// Maximum of supported general performance counter
#define MAX_COUNTER 2
#endif


/**
   \struct InstCounterSettings
   \brief Instruction Counter Settings

   Counter must be set according your intel architecture. For
   example, for Core 2 counter must be zero or one.
*/
struct InstCounterSettings {
  /** General performance counter number */
  int counter;
  /** Enable counter while CPL > 0 (User space) */
  int user;
  /** Enable counter while CPL = 0 (Kernel space) */
  int sys;
};

/**
   \struct FixedCounterSettings
   \brief Fixed Counter Settings
*/
struct FixedCounterSettings {
  /** Enable counter while CPL > 0 (User space) */
  int user;
   /** Enable counter while CPL = 0 (Kernel space) */
  int sys;
};


//===================================================
// Functions
//===================================================

/**
   \brief Disable instruction counter 
   \param[in] s Instruction counter settings
*/
void disableInstructionCounter(struct InstCounterSettings* s);

/**
   \brief Enable instruction counter
   \param[in] s Instruction counter settings
*/
void enableInstructionCounter(struct InstCounterSettings* s);

/**
   \brief Get the performance counter value
   \param[in] s Instruction counter settings
   \return The number of instructions retired
*/
unsigned long long instructionCounter(struct InstCounterSettings* s);

/**
   \brief Setup a counter to count instructions
   
   Set its value to zero

   \param[in] s Instruction counter settings
   \return Zero if successful
 */
unsigned int resetInstructionCounter(struct InstCounterSettings* s);

/**
   \brief Disable CPU core cycles counter
 */
void disableCpuCoreCyclesCounter();

/**
   \brief Enable CPU core cycles counter
   \param[in] s Fixed counter settings
 */
void enableCpuCoreCyclesCounter(struct FixedCounterSettings* s);

/**
   \brief Get the number of cycles since the reset
   \return Number of CPU core cycles
 */
unsigned long long cpuCoreCyclesCounter();

/**
   \brief Reset the CPU core cycles counter
   
   \param[in] s Instruction counter settings
   \return Zero if successful
 */
unsigned int resetCpuCoreCyclesCounter();
